/*
 * lexa-table.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>

#include "lexa-table.h"

#define DEBUG 0
#include "debug.h"


LEXA_TABLE *lexa_table_init(int init_size, int empty_state_offset) {
	LEXA_TABLE *table;

	assert(init_size > 1);
	assert(empty_state_offset >= 1);

	table = (LEXA_TABLE *) malloc(sizeof(LEXA_TABLE));
	if (!table)
		return NULL;

	table->tr_size = init_size;
	table->tr_count = 0;
	table->tr = (LEXA_TABLE_TRANS *) malloc(sizeof(LEXA_TABLE_TRANS) * init_size);

	if (!table->tr) {
		lexa_table_celanup(table);
		return NULL;
	}

	lexa_table_add(table, 0, 0, 0, 0, 0);
	table->empty_state = empty_state_offset;
	table->flag.ready = 1;

	return table;
}

/****************************************************************************/

void lexa_table_celanup(LEXA_TABLE *table) {
	assert(table != NULL);
	
	if(table->tr)
		free(table->tr);
	if(table)
		free(table);
}

/****************************************************************************/

LEXA_TABLE_TRANS *lexa_table_add(LEXA_TABLE *table, int state, unsigned char character, int next_state, int finish_type, int priority) {
	LEXA_TABLE_TRANS *t;
	size_t ns;

	assert(table != NULL);
	assert(state >= 0);

	DBG_MSG(6, "lexa_table_add '%c' (0x%02X) to position %d/%d", isprint(character) ? character : '?', character, table->tr_count, table->tr_size);
	if (table->tr_count >= table->tr_size) {
		ns = table->tr_size * 2;
		DBG_MSG(4, "resizing from %ld to %ld...", (long)table->tr_size, (long)ns);
		t = realloc(table->tr, ns * sizeof(LEXA_TABLE_TRANS));
		if (t) {
			table->tr_size = ns;
			table->tr = t;
		} else {
			perror("lexa_table_add");
			return 0;
		}
	}

	table->tr[table->tr_count].state = state;
	table->tr[table->tr_count].character = character;
	table->tr[table->tr_count].next_state = next_state;
	table->tr[table->tr_count].finish_type = finish_type;
	table->tr[table->tr_count].priority = priority;
	table->flag.ready = 0;
	if (next_state > table->empty_state)
		table->empty_state = next_state +  1;
	else
		table->empty_state++;
	return &(table->tr[table->tr_count++]);
}

/****************************************************************************/

static int lexa_table_trans_compare(const void *a, const void *b) {
	LEXA_TABLE_TRANS *aa = (LEXA_TABLE_TRANS *)a;
	LEXA_TABLE_TRANS *bb = (LEXA_TABLE_TRANS *)b;

	return (aa->state == bb->state) ? 
			(aa->character == bb->character ?
				aa->priority - bb->priority
				: aa->character - bb->character)
			: aa->state - bb->state;
}

LEXA_TABLE_TRANS *lexa_table_lookup(LEXA_TABLE *table, int state, unsigned char ch) {
	LEXA_TABLE_TRANS key;
	int i;
	int sort_again = 0;

	assert(table != NULL);
	assert(state >= 0);

	if (!table->flag.ready) {
		/* sort array */
		qsort(table->tr, table->tr_count, sizeof(LEXA_TABLE_TRANS), lexa_table_trans_compare);
#if 1
		/* remove unused transitions */
		for (i = 1; i < table->tr_count; i++) {
			if (table->tr[i].state == table->tr[i - 1].state && table->tr[i].character == table->tr[i - 1].character) {
				table->tr[i - 1].state = -1;
				table->tr[i].priority = 0;
				sort_again = 1;
			} else {
				table->tr[i - 1].priority = 0;
			}
		}

		/* sort array */
		if (sort_again)
			qsort(table->tr, table->tr_count, sizeof(LEXA_TABLE_TRANS), lexa_table_trans_compare);
#endif		
		/* mark as ready */
		table->flag.ready = 1;
	}

	key.state = state;
	key.character = ch;
	key.priority = 0;
	return bsearch(&key, table->tr, table->tr_count, sizeof(LEXA_TABLE_TRANS), lexa_table_trans_compare);
}

LEXA_TABLE_TRANS *lexa_table_lookup_nosort(LEXA_TABLE *table, int state, unsigned char ch) {
	int i, max_priority = 0, max = -1;
	

	assert(table != NULL);
	assert(state >= 0);

	for (i = 0; i < table->tr_count; i++) {
		if (state == table->tr[i].state && ch == table->tr[i].character) {
			if (max_priority <= table->tr[i].priority) {
				max_priority = table->tr[i].priority;
				max = i;
			}
		}
	}
	if (max == -1) {
		return NULL;
	} else {
		return &(table->tr[max]);
	}
}

/****************************************************************************/

void lexa_table_dump(LEXA_TABLE *table, FILE *file) {
	long i;

	assert(table != NULL);
	assert(file != NULL);

	fprintf(file, "  state | char | priority | next state | finish type  \n");
	fprintf(file, " -------+------+----------+------------+--------------\n");
	for (i = 0; i < table->tr_count; i++) {
		fprintf(file, "  %5d |   %c  | %8d | %10d | %11d\n",
				table->tr[i].state,
				isprint(table->tr[i].character) ? table->tr[i].character : '?',
				table->tr[i].priority,
				table->tr[i].next_state,
				table->tr[i].finish_type
		);
	}
	fprintf(file, " -------+------+----------+------------+--------------\n");
	fprintf(file, "  Size: %ld lines, %ld of %ld KB used\n", i, i * sizeof(LEXA_TABLE_TRANS) >> 10,
			(long) table->tr_size * sizeof(LEXA_TABLE_TRANS) >> 10);
}


