/*
 * lexa.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *
 * TODO
 *
 * 	* Change ranges of characters to bitmaps
 * 	* Write functions for easy creation of state tranition table
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>

#include "lexa.h"
#include "lexa-table.h"

#define DEBUG 0
#include "debug.h"


static int lexa_getc(LEXA *lexa) {
	int c;

	c = getc(lexa->file);
	lexa->byte++;
	if (c == '\n')
		lexa->line++;
	return c;
}

static int lexa_ungetc(LEXA *lexa, int c) {
	if (c == '\n')
		lexa->line--;
	lexa->byte--;
	return ungetc(c, lexa->file);
}


/****************************************************************************/

LEXA *lexa_init(FILE *file, int empty_state_offset) {
	LEXA *lexa;

	assert(file != NULL);

	lexa = (LEXA *) malloc(sizeof(LEXA));
	if (!lexa)
		return NULL;

	lexa->file = file;
	lexa->file_name = NULL;
	lexa->flag.was_opened = 0;

	lexa->table = lexa_table_init(128, empty_state_offset);

	lexa->line = 0;
	lexa->byte = 0;
	lexa->buf = NULL;
	lexa->buf_len = 0;

	return lexa;
}

/****************************************************************************/

LEXA *lexa_open(const char *filename, int empty_state_offset) {
	FILE *f;
	LEXA *lexa;

	assert(filename != NULL);

	f = fopen(filename, "rt");
	if (!f)
		return NULL;

	lexa = lexa_init(f, empty_state_offset);

	lexa->file_name = strdup(filename);
	lexa->flag.was_opened = 1;

	return lexa;
}

/****************************************************************************/

int lexa_cleanup(LEXA *lexa) {
	int ret = 0;

	assert(lexa != NULL);

	if(lexa->table)
		lexa_table_celanup(lexa->table);

	if (lexa->flag.was_opened) {
		if (fclose(lexa->file) == EOF)
			ret = errno;
	}

	if (lexa->file_name) {
		free(lexa->file_name);
	}

	if (lexa->buf) {
		free(lexa->buf);
	}

	free(lexa);

	return ret;
}

/****************************************************************************/

int lexa_add_tr_p(LEXA *lexa, int state, unsigned char char_min, unsigned char char_max,
		unsigned char *char_exclude, unsigned char *char_enum, int next_state, int finish_type, int priority) {
	
	unsigned char *p;
	int ch;

	assert(lexa != NULL);
	assert(char_min <= char_max);

	if (char_enum) {
		DBG_MSG(3, "lexa_add_tr: Adding \"%s\"", char_enum);
		for (p = char_enum; *p; p++) {
			if(!lexa_table_add(lexa->table, state, *p, next_state, finish_type, priority))
				return 0;
		}
	} else if (char_min <= char_max) {
		DBG_MSG(3, "lexa_add_tr: Adding '%c' (0x%02X) ... '%c' (0x%02X)",
				isprint(char_min) ? char_min : '?', char_min,
				isprint(char_max) ? char_max : '?', char_max);
		for (ch = char_min; ch <= char_max; ch++) {
			if (char_exclude == NULL || !strchr(char_exclude, ch)) {
				if (!lexa_table_add(lexa->table, state, (unsigned char)ch, next_state, finish_type, priority - 1))
					return 0;
			}
		}
	}
	DBG_MSG(3, "lexa_add_tr: Done.");
	return 1;
}

/****************************************************************************/

int lexa_add_string_wf(LEXA *lexa, char *str, int finish_type, char *fallback, int fallback_state, int fallback_type) {
	char *p_str = str;
	LEXA_TABLE_TRANS *tr = NULL, *last_tr = NULL;
	int cur_state = 0;
	int i;
	int priority = 3;

	DBG_MSG(3, "lexa_add_string \"%s\"", str);
	assert(lexa != NULL);
	assert(str != NULL);


	/* Build new subtree (if does not exist yet) */
	while (*p_str) {
		tr = lexa_table_lookup_nosort(lexa->table, cur_state, *p_str);
		if (tr && tr->priority == priority) {
			/* character exists - switch to new state only */
			cur_state = tr->next_state;
			last_tr = tr;
		} else {
			/* add new transition and state */
			last_tr = lexa_table_add(lexa->table, cur_state, *p_str, lexa->table->empty_state, fallback_type, priority);
			cur_state = last_tr->next_state;

			if (fallback) {
				lexa_add_tr_p(lexa, cur_state, 0, 0, NULL, fallback, fallback_state, fallback_type, 0);
			}
		}
		p_str++;
	}

	/* check end store finish type */
	if (last_tr) {
		if (last_tr->finish_type > 0 && last_tr->finish_type != finish_type && last_tr->finish_type != fallback_type) {
			fprintf(stderr, "Lexa WARNING: Type collision when building table from string \"%s\". Type %d overrides type %d !\n",
					str, finish_type, last_tr->finish_type);
		}

		for (i = 0; i < lexa->table->tr_count; i++) {
			if (lexa->table->tr[i].next_state == last_tr->next_state)
				lexa->table->tr[i].finish_type = finish_type;
		}
		last_tr->finish_type = finish_type;
	}
	return 1;
}

/****************************************************************************/

char *lexa_scan(LEXA *lexa, int *type) {
	int c;
	int state = 0, first = 1;
	int error = 0;
	char buf[1024];
	char *p_buf = buf;
	LEXA_TABLE_TRANS *tr = NULL;
	LEXA_TABLE_TRANS *last_tr = NULL;

	assert(lexa != NULL);
	assert(lexa->table != NULL);
	assert(lexa->file != NULL);

	if (feof(lexa->file)) {
		*type = 0;
		return NULL;
	}

	if (lexa->line == 0)
		lexa->line = 1;
	lexa->token_byte = lexa->byte;
	lexa->token_line = lexa->line;

	do {
		/* read one character */
		c = lexa_getc(lexa);
		DBG_MSG(6, "State: %d, byte '%c'", state, c);

		/* find next state */
		tr = lexa_table_lookup(lexa->table, state, c);
		if ((!error && tr) || (error && !tr)) {
			*p_buf++ = c;
			*p_buf = 0;
			if (tr)
				state = tr->next_state;
		} else {
			if (first) {
				*p_buf++ = c;
				*p_buf = 0;
				error = 1;
			} else {
				lexa_ungetc(lexa, c);
				*type = error ? -1 : (last_tr ? last_tr->finish_type : 0);
				return strdup(buf);
			}
		}
		last_tr = tr;
		first = 0;
	} while (!feof(lexa->file));

	*type = last_tr ? last_tr->finish_type : 0;
	return strdup(buf);
}

/****************************************************************************/

long lexa_get_line(LEXA *lexa) {
	assert(lexa != NULL);
	return lexa->token_line;
}

/****************************************************************************/

long lexa_get_byte(LEXA *lexa) {
	assert(lexa != NULL);
	return lexa->token_byte;
}

