/*
 * lexa.h
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef JK_LEXA_H
#define JK_LEXA_H

#include <stdio.h>

#include "lexa-table.h"


/* Lexa's main data structure - do not acces this directly */
typedef struct {
	FILE *file;			/* input stream */
	char *file_name;		/* input file name (for info only) */
	long line, byte;		/* position in file */
	long token_line, token_byte;	/* position of last token */
	LEXA_TABLE *table;		/* state transition table */

	struct {
		unsigned was_opened :1;	/* lexa opened file itself ?  */
		unsigned new_line :1;
	} flag;

	char *buf;			/* input buffer */
	size_t buf_len;			/* current input buffer size */
} LEXA;


/* lexa initialization, use opened file stream
 *
 * returns NULL if error
 */
LEXA *lexa_init(FILE *file, int empty_state_offset);


/* lexa initialization, open file itself
 *
 * returns NULL if error
 */
LEXA *lexa_open(const char *filename, int empty_state_offset);


/* cleanup - release memory, close input file if
 *
 * returns zero if succes
 */
int lexa_cleanup(LEXA *lexa);


/* add transition into lexa's state transition table
 *
 * returns 1 if success, zero if failed
 */
#define lexa_add_tr(lexa, state, char_min, char_max, char_exclude, char_enum, next_state, finish_type) \
	lexa_add_tr_p((lexa), (state), (char_min), (char_max), (char_exclude), (char_enum), (next_state), (finish_type), 3)

int lexa_add_tr_p(LEXA *lexa, int state, unsigned char char_min, unsigned char char_max,
		unsigned char *char_exclude, unsigned char *char_enum, int next_state, int finish_type, int priority);

/* convert string to transitions and add it into lexa's state transition table
 *
 * returns 1 if success, zero if failed
 */
#define lexa_add_string(lexa, str, finish_type) \
	lexa_add_string_wf((lexa), (str), (finish_type), NULL, 0, 0)

int lexa_add_string_wf(LEXA *lexa, char *str, int finish_type, char *fallback, int fallback_state, int fallback_type);

/* read one token
 *
 * arguments:
 * 	token:	value of read token, pointer to string (char*)
 * 		On error, token = NULL; otherwise points to strdup(lexa->buf)
 * 		Must be freed after use.
 *
 * returns token's type or zero when error or eof is detected
 */
char *lexa_scan(LEXA *lexa, int *type);


/* get actual line in parsed file
 *
 * returns line number (starting from 1)
 */
long lexa_get_line(LEXA *lexa);


/* get actual column at parsed line
 *
 * returns actual column number (starting from 0)
 */
long lexa_get_byte(LEXA *lexa);


#endif

