/*
 * tree.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "tree.h"

#define DEBUG 0
#include "debug.h"

TREE *tree_new(void)
{
	TREE *tree;

	tree = (TREE *) malloc(sizeof(TREE));
	tree->root = NULL;
	
	return tree;
}


void tree_destroy(TREE* tree)
{
	TREE_ITER *iter;
	assert(tree != NULL);

	iter = tree_iter_new(tree);
	tree_delete(iter);
	tree_iter_destroy(iter);
	free(tree);
	
	return;
}


TREE_NODE *tree_node_new(void)
{
	TREE_NODE *node;

	node = (TREE_NODE *) malloc(sizeof(TREE_NODE));
	node->next = NULL;
	node->prev = NULL;
	node->parent = NULL;
	node->child = NULL;
	node->depth = 0;
	node->token = NULL;
	node->token_type = 0;

	return node;
}

void tree_node_destroy(TREE_NODE *node)
{
	assert(node != NULL);

	if (node->token) {
		free(node->token);
	}
	free(node);
	
	return;
}


TREE_ITER *tree_iter_new(TREE *tree)
{
	TREE_ITER *iter;
	assert(tree != NULL);
	
	iter = (TREE_ITER *) malloc(sizeof(TREE_ITER));
	iter->tree = tree;
	iter->node = tree->root;
	
	return iter;
}


void tree_iter_destroy(TREE_ITER *iter)
{
	assert(iter != NULL);
	
	free(iter);

	return;
}


TREE_ITER *tree_iter_dup(TREE_ITER *iter)
{
	TREE_ITER *new_iter;
	assert(iter != NULL);
	assert(iter->tree != NULL);
	
	new_iter = tree_iter_new(iter->tree);
	new_iter->node = iter->node;
	
	return new_iter;
}


const char *tree_get_token(TREE_ITER *iter)
{
	assert(iter != NULL);
	assert(iter->node != NULL);

	return iter->node->token;
}


int tree_get_token_type(TREE_ITER *iter)
{
	assert(iter != NULL);
	assert(iter->node != NULL);

	return iter->node->token_type;
}


void tree_update(TREE_ITER *iter, const char *token, int token_type)
{
	assert(iter != NULL);
	assert(iter->node != NULL);

	if (iter->node->token)
		free(iter->node->token);

	iter->node->token = (token != NULL ? strdup(token) : NULL);
	iter->node->token_type = token_type;

	return;
}


int tree_append(TREE_ITER *iter)
{
	TREE_NODE *new_node;
	
	assert(iter != NULL);
	assert(iter->tree != NULL);
	assert(iter->tree->root == NULL || iter->node != NULL);
	
	/* create node */
	new_node = tree_node_new();
	if (!new_node)
		return 0;


	/* link node in */
	if (iter->node == NULL && iter->tree->root == NULL) {
		new_node->depth = 0;
		iter->tree->root = new_node;
	} else {
		new_node->depth = iter->node->depth;
		new_node->parent = iter->node->parent;
		new_node->prev = iter->node;
		new_node->next = iter->node->next;
		iter->node->next = new_node;
		if (new_node->next)
			new_node->next->prev = new_node;
	}

	/* move iterator to new node */
	iter->node = new_node;

	return 1;
}

int tree_prepend(TREE_ITER *iter)
{
	TREE_NODE *new_node;
	
	assert(iter != NULL);
	assert(iter->node != NULL || iter->tree->root == NULL);
	
	/* create node */
	new_node = tree_node_new();
	if (!new_node)
		return 0;

	/* link node in */
	if (iter->node == NULL && iter->tree->root == NULL) {
		new_node->depth = 0;
		iter->tree->root = new_node;
	} else {
		new_node->depth = iter->node->depth;
		new_node->parent = iter->node->parent;
		new_node->prev = iter->node->prev;
		new_node->next = iter->node;
		iter->node->next = new_node;
		if (new_node->prev)
			new_node->prev->next = new_node;
	}

	/* move iterator to new node */
	iter->node = new_node;

	return 1;
}


int tree_add_child(TREE_ITER *iter)
{
	TREE_NODE *new_node;
	
	assert(iter != NULL);
	assert(iter->node != NULL || iter->tree->root == NULL);
	assert(iter->node->child == NULL);

	new_node = tree_node_new();
	new_node->parent = iter->node;

	if (iter->node == NULL && iter->tree->root == NULL) {
		new_node->depth = 0;
		iter->tree->root = new_node;
	} else {
		new_node->depth = iter->node->depth + 1;
		iter->node->child = new_node;
	}

	iter->node = new_node;

	return 1;
}

/* delete node and it's childs */
int tree_delete(TREE_ITER *iter)
{
	TREE_ITER *child;
	TREE_NODE *next_node;
	
	assert(iter != NULL);
	assert(iter->node != NULL);

	/* destroy child and it's childs and it's childs' childs and ... */
	child = tree_iter_dup(iter);
	if (tree_iter_move_child(child)) {
		while (tree_delete(child))
			;
	}
	tree_iter_destroy(child);

	/* destroy current node */
	next_node = iter->node->next;
	if (iter->node->prev)
		iter->node->prev->next = iter->node->next;
	if (iter->node->next)
		iter->node->next->prev = iter->node->prev;
	if (iter->node->parent && iter->node->parent->child == iter->node)
		iter->node->parent->child = iter->node->next;
	tree_node_destroy(iter->node);
	iter->node = next_node;
	
	return iter->node != NULL;
}


int tree_iter_valid(TREE_ITER *iter)
{
	assert(iter != NULL);

	return iter->node != NULL;
}


int tree_iter_move_next(TREE_ITER *iter)
{
	assert(iter != NULL);

	if (iter->node)
		iter->node = iter->node->next;
	return iter->node != NULL;
}


int tree_iter_move_prev(TREE_ITER *iter)
{
	assert(iter != NULL);

	if (iter->node)
		iter->node = iter->node->prev;
	return iter->node != NULL;
}


int tree_iter_move_child(TREE_ITER *iter)
{
	assert(iter != NULL);

	if (iter->node)
		iter->node = iter->node->child;
	return iter->node != NULL;
}


int tree_iter_move_parent(TREE_ITER *iter)
{
	assert(iter != NULL);

	if (iter->node)
		iter->node = iter->node->parent;
	return iter->node != NULL;
}


int tree_iter_move_next_linear(TREE_ITER *iter)
{
	assert(iter != NULL);

	if (iter->node == NULL)
		return 0;

	if (iter->node->child) {
		iter->node = iter->node->child;
		return 1;
	} else if (iter->node->next) {
		iter->node = iter->node->next;
		return 1;
	} else if(iter->node->parent && iter->node->parent->next) {
		iter->node = iter->node->parent->next;
		return 1;
	} else {
		iter->node = NULL;
		return 0;
	}
}


void tree_dump(TREE *tree, FILE *file)
{
	TREE_ITER *iter;

	assert(tree != NULL);
	assert(file != NULL);
	
	fprintf(file, "Tree dump (%p)\n", tree);
	if (tree->root) {
		iter = tree_iter_new(tree);
		while (tree_iter_valid(iter)) {
			tree_dump_iter(iter, file, 0);
			tree_iter_move_next(iter);
		}
		tree_iter_destroy(iter);
	} else {
		fprintf(file, "\t(empty)\n");
	}
	
	return;
}


void tree_dump_iter(TREE_ITER *iter, FILE *file, int depth)
{
	TREE_ITER *child;
	int i;

	assert(iter != NULL);
	assert(file != NULL);

	if (iter->node) {
		for (i = 0; i <= iter->node->depth; i++) {
			fputc('\t', file);
		}
		fprintf(file, "%d: (%d) \"%s\"\n", iter->node->depth, iter->node->token_type, iter->node->token);
	
		child = tree_iter_dup(iter);
		if (tree_iter_move_child(child)) {
			while (tree_iter_valid(child)) {
				tree_dump_iter(child, file, depth + 1);
				tree_iter_move_next(child);
			}
		}
		tree_iter_destroy(child);
	}

	return;
}


